/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import com.thedeanda.lorem.LoremIpsum;

class UserProfileResourceTest {

	private static final String USER_MANAGER_URL = LoremIpsum.getInstance().getUrl();

	@InjectMocks
	private UserProfileResource resource;

	@Mock
	private UserService userService;

	@Mock
	private UserProfileResourceAssembler resourceAssembler;

	@BeforeEach
	void init() {
		resource.userManagerUrl = USER_MANAGER_URL;
	}

	@Nested
	class TestFindById {

		private static final String USER_ID = UUID.randomUUID().toString();
		private final User user = UserTestFactory.create();

		@BeforeEach
		void init() {
			when(userService.getById(anyString())).thenReturn(user);
		}

		@Test
		void shouldCallUserService() {
			resource.findById(USER_ID);

			verify(userService).getById(USER_ID);
		}

		@Test
		void shouldCallResourceAssembler() {
			resource.findById(USER_ID);

			verify(resourceAssembler).toUserProfileResource(user, USER_MANAGER_URL);
		}
	}

	@Nested
	class TestFindUsers {

		private static final String SEARCH_QUERY = "abc";
		private static final List<User> users = List.of(UserTestFactory.create());
		private static Optional<Boolean> EMPTY_PARAM_DELETED = Optional.empty();
		private static Optional<Boolean> DELETED_USER_PROFILES = Optional.of(true);
		private static Optional<Boolean> NOT_DELETED_USER_PROFILES = Optional.of(false);
		private static Optional<String> EMPTY_PARAM_ORGANISATIONSEINHEIT_ID = Optional.empty();

		@Captor
		private ArgumentCaptor<FindUserProfilesQuery> query;

		@Test
		void shouldFindNotDeletedUserProfiles() {
			resource.findUsers(SEARCH_QUERY, NOT_DELETED_USER_PROFILES, EMPTY_PARAM_ORGANISATIONSEINHEIT_ID);

			verify(userService).findActiveUsers(query.capture());
			assertThat(query.getValue()).usingRecursiveComparison().isEqualTo(
					FindUserProfilesQuery
							.builder()
							.searchBy(SEARCH_QUERY)
							.limit(UserProfileResource.LIMIT)
							.build());
		}

		@Test
		void shouldFindDeletedUserProfiles() {
			resource.findUsers(SEARCH_QUERY, DELETED_USER_PROFILES, EMPTY_PARAM_ORGANISATIONSEINHEIT_ID);

			verify(userService).findInactiveUsers(query.capture());
			assertThat(query.getValue()).usingRecursiveComparison().isEqualTo(
					FindUserProfilesQuery
							.builder()
							.searchBy(SEARCH_QUERY)
							.limit(UserProfileResource.LIMIT)
							.build());
		}

		@Test
		void shouldFindAllUserProfiles() {
			resource.findUsers(SEARCH_QUERY, EMPTY_PARAM_DELETED, EMPTY_PARAM_ORGANISATIONSEINHEIT_ID);

			verify(userService).findUsers(query.capture());
			assertThat(query.getValue()).usingRecursiveComparison().isEqualTo(
					FindUserProfilesQuery
							.builder()
							.searchBy(SEARCH_QUERY)
							.limit(UserProfileResource.LIMIT)
							.build());
		}

		@Test
		void shouldCallResourceAssembler() {
			when(userService.findUsers(any())).thenReturn(users.stream());

			resource.findUsers(SEARCH_QUERY, EMPTY_PARAM_DELETED, EMPTY_PARAM_ORGANISATIONSEINHEIT_ID);

			verify(resourceAssembler).toUserProfileResourceList(users, USER_MANAGER_URL);
		}
	}
}