/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user.keycloak;

import static de.ozgcloud.user.UserTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.function.Supplier;
import java.util.stream.Stream;

import jakarta.ws.rs.BadRequestException;
import jakarta.ws.rs.ClientErrorException;
import jakarta.ws.rs.NotFoundException;
import jakarta.ws.rs.ProcessingException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.keycloak.admin.client.resource.RealmResource;
import org.keycloak.admin.client.resource.UserResource;
import org.keycloak.admin.client.resource.UsersResource;
import org.keycloak.representations.idm.UserRepresentation;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.user.RemoteUserIterator;
import de.ozgcloud.user.User;
import de.ozgcloud.user.UserRepresentationTestFactory;
import de.ozgcloud.user.UserResourceMapper;
import de.ozgcloud.user.UserTestFactory;
import de.ozgcloud.user.common.errorhandling.KeycloakUnavailableException;

class KeycloakApiServiceTest {

	@Spy
	@InjectMocks
	private KeycloakApiService service;
	@Mock
	private RealmResource realmResource;
	@Mock
	private UserResourceMapper userResourceMapper;
	@Mock
	private KeycloakApiProperties properties;

	@Mock
	private UserResource userResource;

	@DisplayName("Find all user")
	@Nested
	class TestFindAllUser {

		@Mock
		private UsersResource usersResource;

		@Mock
		private RemoteUserIterator iterator;

		private final User user = UserTestFactory.create();
		private final UserRepresentation userRepresentation = UserRepresentationTestFactory.create();
		private final Stream<UserRepresentation> userStream = Stream.of(userRepresentation);

		@BeforeEach
		void init() {
			doReturn(userStream).when(service).getAllUserRepresentation();
		}

		@Test
		void shouldCreateIteratorWithRealmResource() {
			service.findAllUser();

			verify(service).getAllUserRepresentation();
		}

		@Test
		void shouldCallMapper() {
			when(userResourceMapper.toKopUser(any(UserResource.class))).thenReturn(user);
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.get(anyString())).thenReturn(userResource);

			service.findAllUser().toList();

			verify(userResourceMapper).toKopUser(any(UserResource.class));
		}
	}

	@Nested
	@DisplayName("Find user by id")
	class TestFindUserById {

		private static final String USER_ID = "user_id";

		@Mock
		private UsersResource usersResource;

		@BeforeEach
		void beforeEach() {
			when(realmResource.users()).thenReturn(usersResource);
		}

		@Test
		void shouldCallUsersResource() {
			when(usersResource.get(anyString())).thenReturn(userResource);

			service.findUserById(EXTERNAL_ID);

			verify(usersResource).get(EXTERNAL_ID);
		}

		@Test
		void shouldCallMapper() {
			when(usersResource.get(anyString())).thenReturn(userResource);

			service.findUserById(EXTERNAL_ID);

			verify(userResourceMapper).toKopUser(any());
		}

		@Test
		@DisplayName("should not call mapper when user not found")
		void shouldNotCallMapper() {
			when(usersResource.get(anyString())).thenThrow(new NotFoundException());

			service.findUserById("unknown_user");

			verify(userResourceMapper, never()).toKopUser(any());
		}

		@Test
		void shoulReturnEmpty() {
			when(usersResource.get(USER_ID)).thenThrow(new NotFoundException());

			var user = service.findUserById(USER_ID);

			assertThat(user).isEmpty();
		}
	}

	@DisplayName("handling keycloak exception")
	@Nested
	class TestHandlingKeycloakException {

		@Test
		void shouldThrowOnClientErrorException() {
			assertThatThrownBy(() -> handlingKeyCloakException(() -> {
				throw new ClientErrorException(403);
			})).isInstanceOf(KeycloakUnavailableException.class);
		}

		@Test
		void shouldThrowOnProcessingException() {
			assertThatThrownBy(() -> handlingKeyCloakException(() -> {
				throw new ProcessingException(new RuntimeException());
			})).isInstanceOf(KeycloakUnavailableException.class);
		}

		@Test
		void shouldThrowIllegalStateException() {
			assertThatThrownBy(() -> handlingKeyCloakException(() -> {
				throw new IllegalStateException(new RuntimeException());
			})).isInstanceOf(KeycloakUnavailableException.class);
		}

		private <T> T handlingKeyCloakException(Supplier<T> runnable) {
			return service.handlingKeycloakException(runnable);
		}
	}

	@DisplayName("Set user attribute")
	@Nested
	class TestUpdateUserAttribute {

		private static final String KEYCLOAK_USER_ID = "1ae2-b123";
		private static final String ATTRIBUTE_NAME_USER_ID = "userId";
		private static final String ATTRIBUTE_VALUE_USER_ID = "10";
		private static final String NEW_USER_ID = "1010";

		@Mock
		private UsersResource usersResource;
		@Mock
		private UserResource userResource;
		@Mock
		private UserRepresentation userRepresentation;

		@BeforeEach
		void beforeEach() {
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.get(KEYCLOAK_USER_ID)).thenReturn(userResource);
			when(userResource.toRepresentation()).thenReturn(userRepresentation);
			when(userRepresentation.firstAttribute(ATTRIBUTE_NAME_USER_ID)).thenReturn(ATTRIBUTE_VALUE_USER_ID);
		}

		@Test
		void shouldGetUsersResource() {
			service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);

			verify(realmResource).users();
		}

		@Test
		void shouldGetUserResource() {
			service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);

			verify(usersResource).get(KEYCLOAK_USER_ID);
		}

		@Test
		void shouldGetUserRepresentation() {
			service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);

			verify(userResource).toRepresentation();
		}

		@Test
		void shouldUpdateAttributeIfNotExist() {
			when(userRepresentation.firstAttribute(ATTRIBUTE_NAME_USER_ID)).thenReturn(null);

			service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);

			verify(userRepresentation).singleAttribute(ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);
			verify(service).tryUpdateUserResource(userResource, userRepresentation, ATTRIBUTE_NAME_USER_ID);
		}

		@Test
		void shouldUpdateAttributeIfExistsWithDifferentValue() {
			when(userRepresentation.firstAttribute(ATTRIBUTE_NAME_USER_ID)).thenReturn(NEW_USER_ID);

			service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);

			verify(userRepresentation).singleAttribute(ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);
			verify(service).tryUpdateUserResource(userResource, userRepresentation, ATTRIBUTE_NAME_USER_ID);
		}

		@Test
		void shouldUpdateAttributeIfNewOneGiven() {
			service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, NEW_USER_ID);

			verify(userRepresentation).singleAttribute(ATTRIBUTE_NAME_USER_ID, NEW_USER_ID);
			verify(service).tryUpdateUserResource(userResource, userRepresentation, ATTRIBUTE_NAME_USER_ID);
		}

		@Test
		void shouldNotSetAttributeIfExistsWithSameValue() {
			service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);

			verify(userRepresentation, never()).singleAttribute(ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID);
			verify(service, never()).tryUpdateUserResource(userResource, userRepresentation, ATTRIBUTE_NAME_USER_ID);
		}

		@Test
		void shouldCatchBadRequestException() {
			when(userRepresentation.firstAttribute(ATTRIBUTE_NAME_USER_ID)).thenReturn(NEW_USER_ID);
			doThrow(new BadRequestException("error message")).when(userResource).update(any(UserRepresentation.class));

			assertThatCode(
					() -> service.updateAttribute(KEYCLOAK_USER_ID, ATTRIBUTE_NAME_USER_ID, ATTRIBUTE_VALUE_USER_ID)).doesNotThrowAnyException();
		}
	}

	@Nested
	class TestTryUpdateUserResource {

		private static final String ATTRIBUTE_NAME_USER_ID = "userId";

		@Captor
		private ArgumentCaptor<UserRepresentation> userRepresentationArgumentCaptor;

		@Mock
		private UserRepresentation userRepresentation;

		@Test
		void shouldUpdateUserResource() {
			service.tryUpdateUserResource(userResource, userRepresentation, ATTRIBUTE_NAME_USER_ID);

			verify(userResource).update(any(UserRepresentation.class));
		}

		@Test
		void shouldCreateOlderUserRepresentation() {
			service.tryUpdateUserResource(userResource, UserRepresentationTestFactory.create(), ATTRIBUTE_NAME_USER_ID);

			verify(userResource).update(userRepresentationArgumentCaptor.capture());
			assertThat(userRepresentationArgumentCaptor.getValue()).usingRecursiveComparison().isEqualTo(UserRepresentationTestFactory.create());
		}

	}
}